<?php
include_once(dirname(__FILE__).'../../../config/config.inc.php');
include_once(dirname(__FILE__).'../../../init.php');
include_once(_PS_MODULE_DIR_.'routee/routee.php');
$routeeObj = new Routee();

const ROUTEE_MASS_LOG_URL = 'https://autoforms.routee.net/u001/log.php?multi=1';

switch (Tools::getValue('method')) {
    case 'exportCsv' :
        $return = handleLogsCsv();
        echo json_encode($return);
        break;
    case 'exportLogApi' :
        $response = handleLogsApi();
        echo json_encode($response);
        break;
    case 'productExport' :
        $response = handleProductExport(Tools::getValue('cyclecount'), $routeeObj);
        echo json_encode($response);
        break;
    case 'customerExport' :
        $response = handleCustomerExport(Tools::getValue('cyclecount'), $routeeObj);
        echo json_encode($response);
        break;
    case 'orderExport' :
        $response = handleOrderExport(Tools::getValue('cyclecount'), $routeeObj);
        echo json_encode($response);
        break;
    default:
        exit;
}

/**
 * @param $count
 * @param $routeeObj
 * @return array
 * @throws PrestaShopDatabaseException
 */
function handleProductExport($count, $routeeObj) {
    $routeeObj->eventExecutedLog('MassProduct', 'massdata');

    $uuid = $routeeObj->getUUID();
    $limit = 100;
    $offset = $count * $limit;

    $DBProducts = Db::getInstance()->executeS(
        "SELECT * FROM " . _DB_PREFIX_ . "product ORDER BY id_product LIMIT $limit OFFSET $offset"
    );

    if(!empty($DBProducts)) {
        $routeeObj->eventGrabData('MassProduct', count($DBProducts), 'massdata');

        $products = array();

        foreach ($DBProducts as $product) {
            $products[] = $routeeObj->getProductData($product["id_product"]);
        }

        $prodObject = new stdClass();
        $prodObject->name = "products";
        $prodObject->description = "shop products";
        $prodObject->object = $products;

        $routeeObj->eventPayloadData('MassProduct', $prodObject, 'massdata');

        $routeeObj->sendDbData($prodObject, $uuid);

        if($offset == 0) {
            return array('status' => 'success', 'count' => $count + 1, 'msg' => 'The product export process started.', 'reload' => 0);
        }

        return array('status' => 'success', 'count' => $count + 1, 'msg' => $offset.' Products exported successfully', 'reload' => 0);
    } else {
        Configuration::updateValue('PS_ROUTEE_MASS_PRODUCT', 1);
        return array('status' => 'success', 'count' => $count + 1, 'msg' => $offset.' Products exported successfully', 'reload' => 1);
    }
}

/**
 * @param $count
 * @param $routeeObj
 * @return array
 * @throws PrestaShopDatabaseException
 */
function handleCustomerExport($count, $routeeObj) {
    $routeeObj->eventExecutedLog('MassCustomer', 'massdata');

    $uuid = $routeeObj->getUUID();
    $limit = 100;
    $offset = $count * $limit;
    $cartUrl = Tools::getHttpHost(true) . __PS_BASE_URI__ . "cart?action=show";

    $customers = Db::getInstance()->executeS(
        "SELECT * FROM " . _DB_PREFIX_ . "customer ORDER BY id_customer LIMIT $limit OFFSET $offset"
    );

    if(!empty($customers)) {
        $routeeObj->eventGrabData('MassCustomer', count($customers), 'massdata');

        $returnData = array();
        if ((is_array($customers) || is_object($customers))) {
            foreach ($customers as $customer) {
                $id = $customer["id_customer"];

                //Set customer format
                $customerParsed = array(
                    "id" => $customer["id_customer"],
                    "lastname" => $customer["lastname"],
                    "firstname" => $customer["firstname"],
                    "birthday" => $customer["birthday"],
                    "email" => $customer["email"],
                    "company" => $customer["company"],
                    "newsletter" => ($customer["newsletter"] == false) ? false : true,
                    "website" => $customer["website"],
                    "date_add" => $customer["date_add"],
                    "date_upd" => $customer["date_upd"]
                );
                $addresses = Db::getInstance()->executeS(
                    "SELECT * FROM " . _DB_PREFIX_ . "address WHERE id_customer='$id'"
                );
                $addressesParsed = array();

                //Set addresses format
                foreach ($addresses as $address) {
                    $addressParsed = array(
                        "lastname" => $address["lastname"],
                        "firstname" => $address["firstname"],
                        "address1" => $address["address1"],
                        "address2" => $address["address2"],
                        "postcode" => $address["postcode"],
                        "phone" => $address["phone"],
                        "phone_mobile" => $address["phone_mobile"],
                        "city" => $address["city"],
                        "country" => $routeeObj->getCountryById($address["id_country"]),
                        "company" => $address["company"]
                    );
                    $addressesParsed[] = $addressParsed;
                }
                $carts = Db::getInstance()->executeS(
                    "SELECT * FROM " . _DB_PREFIX_ . "cart WHERE id_customer='$id'"
                );

                $respCartProducts = array();
                $respCartDetails = new stdClass();
                // Fill cart with products
                if (is_array($carts) || is_object($carts)) {
                    foreach ($carts as $cart) {
                        $cartId = $cart['id_cart'];
                        $orders = Db::getInstance()->executeS(
                            "SELECT * FROM " . _DB_PREFIX_ . "orders WHERE id_cart='$cartId'"
                        );
                        $cart_products = Db::getInstance()->executeS(
                            "SELECT * FROM " . _DB_PREFIX_ . "cart_product WHERE id_cart='$cartId'"
                        );
                        if (!(is_array($orders) && sizeof($orders) > 0)) {
                            if (is_array($cart_products) && sizeof($cart_products) > 0) {
                                $respCartDetails->date_add = $cart["date_add"];
                                $respCartDetails->date_upd = $cart["date_upd"];

                                foreach ($cart_products as $cart_product) {
                                    $temp_cart_product = array();
                                    $temp_cart_product["product_id"] = $cart_product->id_product;
                                    $temp_cart_product["option_id"] = $cart_product->id_product_attribute;
                                    $temp_cart_product["quantity"] = $cart_product->quantity;
                                    $temp_cart_product["quantity"] = $cart_product->quantity;
                                    $temp_cart_product["date_add"] = $cart_product->date_add;
                                    $respCartProducts[] = $temp_cart_product;
                                }
                            }
                        }
                    }
                }

                $tempReturnObject = new stdClass();
                $tempReturnObject->customer = $customerParsed;
                $tempReturnObject->addresses = $addressesParsed;
                $tempReturnObject->cart = $respCartDetails;
                $tempReturnObject->cart->cart_products = $respCartProducts;
                $tempReturnObject->cart->cart_url = $cartUrl;
                $returnData[] = $tempReturnObject;
            }
        }
        $custObject = new stdClass();
        $custObject->name = "customers";
        $custObject->description = "shop customers";
        $custObject->object = $returnData;

        $routeeObj->eventPayloadData('MassCustomer', $custObject, 'massdata');

        $routeeObj->sendDbData($custObject, $uuid);

        if($offset == 0) {
            return array('status' => 'success', 'count' => $count + 1, 'msg' => 'The customer export process started.', 'reload' => 0);
        }

        return array('status' => 'success', 'count' => $count + 1, 'msg' => $offset.' Customers exported successfully', 'reload' => 0);
    } else {
        Configuration::updateValue('PS_ROUTEE_MASS_CUSTOMER', 1);
        return array('status' => 'success', 'count' => $count + 1, 'msg' => $offset.' Customers exported successfully', 'reload' => 1);
    }
}

/**
 * @param $count
 * @param $routeeObj
 * @return array
 * @throws PrestaShopDatabaseException
 * @throws PrestaShopException
 */
function handleOrderExport($count, $routeeObj) {
    $routeeObj->eventExecutedLog('MassOrder', 'massdata');

    $uuid = $routeeObj->getUUID();
    $limit = 100;
    $offset = $count * $limit;

    $DBOrders = Db::getInstance()->executeS(
        "SELECT * FROM " . _DB_PREFIX_ . "orders ORDER BY id_order LIMIT $limit OFFSET $offset"
    );

    if(!empty($DBOrders)) {
        $routeeObj->eventGrabData('MassOrder', count($DBOrders), 'massdata');

        $langId = Context::getContext()->context->language->id;
        $orders = array();

        foreach ($DBOrders as $DBOrder) {
            $order = new Order($DBOrder["id_order"], $langId);
            $customer = new Customer($order->getCustomer()->id);

            $orderStatus = $order->getCurrentOrderState();
            $address = $customer->getAddresses($langId)[0];
            $orderProducts = $order->getProducts();
            $orderProductsParsed = array();

            foreach ($orderProducts as $product) {
                $temp = array();
                $temp["product_id"] = $product["product_id"];
                $temp["option_id"] = $product["product_attribute_id"];
                $temp["quantity"] = $product["product_quantity"];
                $orderProductsParsed[] = $temp;
            }
            $orderData = array();
            $orderData["order_id"] = $order->id;
            $orderData["status"] = $orderStatus->name[$langId];
            $orderData["tracking_id"] = $order->shipping_number;
            $orderData["transport_company"] = $routeeObj->getCarrierName($order->id);
            $orderData["shipping_method"] = $routeeObj->getCarrierName($order->id);
            $orderData["payment_method"] = $order->payment;
            $orderData["total_price"] = $order->total_paid_tax_incl;
            $orderData["customer_id"] = $order->id_customer;
            $orderData["birthday"] = $customer->birthday;
            $orderData["email"] = $customer->email;
            $orderData["newsletter"] = $customer->newsletter;
            $orderData["website"] = $customer->website;
            $orderData["company"] = $address["company"];
            $orderData["lastname"] = $address["lastname"];
            $orderData["firstname"] = $address["firstname"];
            $orderData["address1"] = $address["address1"];
            $orderData["address2"] = $address["address2"];
            $orderData["postcode"] = $address["postcode"];
            $orderData["city"] = $address["city"];
            $orderData["country"] = $address["country"];
            $orderData["phone"] = $address["phone"];
            $orderData["phone_mobile"] = $address["phone_mobile"];
            $orderData["date_add"] = $address["date_add"];
            $orderData["date_upd"] = $address["date_upd"];
            $orderData["order_products"] = $orderProductsParsed;
            $orders[] = $orderData;
        }
        $orderObject = new stdClass();
        $orderObject->name = "orders";
        $orderObject->description = "shop orders";
        $orderObject->object = $orders;

        $routeeObj->eventPayloadData('MassOrder', $orderObject, 'massdata');

        $routeeObj->sendDbData($orderObject, $uuid);

        if($offset == 0) {
            return array('status' => 'success', 'count' => $count + 1, 'msg' => 'The Order export process started.', 'reload' => 0);
        }

        return array('status' => 'success', 'count' => $count + 1, 'msg' => $offset.' Orders exported successfully', 'reload' => 0);
    } else {
        Configuration::updateValue('PS_ROUTEE_MASS_ORDER', 1);
        return array('status' => 'success', 'count' => $count + 1, 'msg' => $offset.' Orders exported successfully', 'reload' => 1);
    }
}

/**
 * @return array
 * @throws PrestaShopDatabaseException
 */
function handleLogsCsv()
{
    $response = Db::getInstance()->executeS(
        "SELECT * FROM " . _DB_PREFIX_ . "store_events_logs WHERE is_exported='0'"
    );
    return array('status' => 'success', 'data' => $response, 'msg' => 'Logs CSV downloaded successfully.', 'reload' => 1);
}

/**
 * @return array
 * @throws PrestaShopDatabaseException
 */
function handleLogsApi()
{
    $uuid = Db::getInstance()->executeS("SELECT value FROM `" . _DB_PREFIX_ . "configuration` WHERE name='uuid';");
    $uuid = $uuid[0]['value'] ?? '';


    $offset = $delayCounter = 0;
    $limit = 100;
    $flag = true;
    $postArr = $ids = [];
    while ($flag) {
        $logs = Db::getInstance()->executeS(
            "SELECT * FROM " . _DB_PREFIX_ . "store_events_logs WHERE is_exported='0' LIMIT $limit OFFSET $offset"
        );

        foreach ($logs as $log) {
            $ids[] = $log['id'];
            $postArr[] = array(
                'siteUrl' => $log['store_url'],
                'uuid' => $uuid,
                'event_name' => eventName($log['event_type']),
                'log_type' => $log['log_type'] == 1 ? 'success' : 'failure',
                'log_data' => $log['log_data'],
                'created_at' => $log['created_at'],
                'platform' => "Prestashop"
            );
        }

        if (count($logs) < $limit) {
            $flag = false;
        }
        $offset += $limit;
        if ($delayCounter == 2) {
            $delayCounter = 0;
            sleep(1);
        }
    }

    $response = logRequest(ROUTEE_MASS_LOG_URL, $postArr, 'POST');
    routeeUpdateLogs($response['code'], $ids);

    return array('status' => 'success', 'data' => $response, 'msg' => 'Logs exported to API successfully.', 'reload' => 1);
}

/**
 * @param  $url
 * @param  $data
 * @param  $method
 * @return array
 */
function logRequest($url, $data, $method)
{
    $curl = curl_init();
    curl_setopt_array($curl, array(
        CURLOPT_URL => $url,
        CURLOPT_RETURNTRANSFER => true,
        CURLOPT_ENCODING => "",
        CURLOPT_MAXREDIRS => 6,
        CURLOPT_TIMEOUT => 30,
        CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
        CURLOPT_CUSTOMREQUEST => $method,
        CURLOPT_HTTPHEADER => array("content-type: application/json"),
        CURLOPT_POSTFIELDS => json_encode($data)
    ));

    $curl_response = curl_exec($curl);
    $error = curl_error($curl);
    $code = curl_getinfo($curl, CURLINFO_HTTP_CODE);
    curl_close($curl);

    $response = array();
    $response["code"] = $code;
    $response["error"] = $error;
    $response["response"] = json_decode($curl_response);
    return $response;
}

/**
 * @param $eventType
 * @return string
 */
function eventName($eventType) {
    $name = '';

    switch ($eventType) {
        case 1:
            $name = "auth";
            break;
        case 2:
            $name = "events";
            break;
        case 3:
            $name = "massdata";
            break;
    }

    return $name;
}

/**
 * @param $code
 * @return void
 */
function routeeUpdateLogs($code, $ids) {
    if ($code == 200) {
        Db::getInstance()->executeS(
            "UPDATE " . _DB_PREFIX_ . "store_events_logs SET is_exported=1 WHERE is_exported=0 AND id IN (".implode(',',$ids).")"
        );
    }
}
